(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-Clang                                      *)
(*                                                                        *)
(*  Copyright (C) 2012-2016                                               *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(*                                                                        *)
(**************************************************************************)

open Cil_types
open Logic_ptree
open Intermediate_format

let const_pkind = function
  | PDataPointer { qualifier } -> List.mem Const qualifier
  | PFunctionPointer _ 
  | PStandardMethodPointer _ 
  | PVirtualMethodPointer _ -> false

let const_valid loc pkind t =
  if const_pkind pkind then
    { lexpr_node = PLvalid_read(None,t); lexpr_loc = loc }
  else { lexpr_node = PLvalid(None,t); lexpr_loc = loc }

let add_valid_result env typ =
  let loc = Convert_env.get_loc env in
  let rec aux = function
    | LVReference kind | RVReference kind ->
        [Cil_types.Normal,
         const_valid loc kind { lexpr_node = PLresult; lexpr_loc = loc}]
    | Named (name,_) when Cxx_utils.is_builtin_qual_type name -> []
    | Named (name,_) -> aux (Convert_env.get_typedef env name).plain_type
    | _ -> []
  in aux typ.plain_type

let add_valid_reference env acc arg =
  let rec aux =
    function
      | LVReference kind | RVReference kind ->
        const_valid arg.arg_loc kind
          { lexpr_node = PLvar(arg.arg_name); lexpr_loc = arg.arg_loc; }
        :: acc
      | Named (name,_) when Cxx_utils.is_builtin_qual_type name -> acc
      | Named (name,_) -> aux (Convert_env.get_typedef env name).plain_type
      | _ -> acc
  in aux arg.arg_type.plain_type

let add_valid_references env l = List.fold_left (add_valid_reference env) [] l

let add_valid_this env =
  let loc = Convert_env.get_loc env in
  function
    | FKCastMethodOperator (cv,_) when List.mem Const cv ->
      [{ lexpr_node =
          PLvalid_read (None, { lexpr_node = PLvar("this"); lexpr_loc = loc});
         lexpr_loc = loc}]
    | FKMethod cv when List.mem Const cv ->
      [{ lexpr_node =
          PLvalid_read (None, { lexpr_node = PLvar("this"); lexpr_loc = loc});
         lexpr_loc = loc}]
    | FKConstructor _ | FKDestructor _ | FKMethod _ | FKCastMethodOperator _ ->
      [ { lexpr_node =
          PLvalid( None, { lexpr_node = PLvar("this"); lexpr_loc = loc });
          lexpr_loc = loc} ]
    | FKFunction -> []

let add_separated_arg env acc arg =
  let (current_class,tc) =
    Convert_env.class_name_from_qualifications env
      (Convert_env.get_namespace env)
  in
  let rec check_class typ =
    match typ, Convert_env.current_struct_or_union env with
      | (Struct (name,ts), (CClass | CStruct)) | Union (name,ts), CUnion
        when Fclang_datatype.Qualified_name.equal (name,ts) (current_class,tc)
        -> { lexpr_node =
            PLseparated(
              [ { lexpr_node = PLvar "this"; lexpr_loc = arg.arg_loc };
                { lexpr_node = PLvar arg.arg_name; lexpr_loc = arg.arg_loc } ]);
          lexpr_loc = arg.arg_loc }
        :: acc
      | Named (name,_), _ when Cxx_utils.is_builtin_qual_type name -> acc
      | Named (name,_), _
        -> check_class (Convert_env.get_typedef env name).plain_type
      | _ -> acc
  in
  let rec aux = function
    | LVReference (PDataPointer t)
    | RVReference (PDataPointer t)
    | Pointer (PDataPointer t) ->
      check_class t.plain_type
    | Named (name,_) when Cxx_utils.is_builtin_qual_type name -> acc
    | Named (name,_) -> aux (Convert_env.get_typedef env name).plain_type
    | _ -> acc
  in aux arg.arg_type.plain_type

let add_separated_constructor env args =
  function
    | FKConstructor _ ->
      List.fold_left (add_separated_arg env) [] args
    | FKDestructor _ | FKMethod _ | FKCastMethodOperator _
    | FKFunction -> []

let add_contract
    ~env ~kind ~return_type ~args ~variadic ~implicit ~extern_c spec
    =
  (* not used yet *)
  let () = ignore (variadic,implicit,extern_c) in
  let requires = add_valid_references env args in
  let post_cond = add_valid_result env return_type in
  let requires = add_valid_this env kind @ requires in
  let requires = add_separated_constructor env args kind @ requires in
  if requires <> [] || post_cond <> [] then begin
    match spec with
      | Some (s,_) ->
        (try
           let default = List.find Cil.is_default_behavior s.spec_behavior in
           default.b_requires <- requires @ default.b_requires;
           default.b_post_cond <- post_cond @ default.b_post_cond;
           spec
         with Not_found ->
           s.spec_behavior <-
             Cil.mk_behavior ~requires ~post_cond () :: s.spec_behavior;
           spec)
      | None ->
        Some
          ({ spec_behavior = [ Cil.mk_behavior ~requires ~post_cond () ];
             spec_variant = None;
             spec_terminates = None;
             spec_complete_behaviors = [];
             spec_disjoint_behaviors = [] },
           Convert_env.get_loc env)
  end else spec
